#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import secrets
import json
import tornado.web
from .utilities import singleton
from .cacheproxy import CacheProxy

_all_user_session_info = {}

@singleton
class HandlerSessionDataManager():
    
    def __init__(self):
        self.session_data_map = {}
        self.cache_path = 'data/session-data.bin'
        if os.path.exists(self.cache_path):
            with open(self.cache_path, 'r+') as f:
                cont = f.read()
                data = json.loads(cont)
                for k, value in data.items():
                    self.session_data_map[k] = value
                f.close()

    def set_item(self, key: str, name: str, value: any):
        # CacheProxy().
        if key not in self.session_data_map:
            self.session_data_map[key] = {}
        self.session_data_map[key][name] = value
        self.save_to_cache()
    
    def get_item(self, key: str, name: str):
        data = self.session_data_map.get(key, None)
        if data:
            return data.get(name, None)
        return None
    
    def key_exists(self, key: str):
        return self.session_data_map.get(key, None)
    
    def delete(self, key: str):
        if key in self.session_data_map:
            del self.session_data_map[key]
        self.save_to_cache()

    def save_to_cache(self):
        data = json.dumps(self.session_data_map)
        if not os.path.exists(self.cache_path):
            path_name, _ = os.path.split(self.cache_path)
            if not os.path.exists(path_name):
                os.makedirs(path_name, 0o777)
        with open(self.cache_path, 'w+') as f:
            f.write(data)
            f.close()

_session_manager = HandlerSessionDataManager()

SESSION_KEY = '__session__'

class TornadoSession(object):
    def __init__(self, handler: tornado.web.RequestHandler):
        self.handler = handler
        self.random_index_str = None

    def __get_random_str(self):
        return secrets.token_urlsafe()

    def __setitem__(self, key, value):
        if not self.random_index_str:
            random_index_str = self.handler.get_secure_cookie(SESSION_KEY, None)
            if random_index_str:
                if not _session_manager.key_exists(random_index_str):
                    self.random_index_str = self.__get_random_str()
            else:
                self.random_index_str = self.__get_random_str()
                self.handler.set_secure_cookie(SESSION_KEY, self.random_index_str)
        _session_manager.set_item(self.random_index_str, key, value)
        # if not self.random_index_str:
        #     random_index_str = self.handler.get_secure_cookie(SESSION_KEY, None)
        #     if random_index_str:
        #         if random_index_str not in _all_user_session_info:
        #             self.random_index_str = self.__get_random_str()
        #             _all_user_session_info[self.random_index_str] = {}
        #     else:
        #         self.random_index_str = self.__get_random_str()
        #         self.handler.set_secure_cookie(SESSION_KEY, self.random_index_str)
        #         _all_user_session_info[self.random_index_str] = {}
        # _all_user_session_info[self.random_index_str][key] = value
        self.handler.set_secure_cookie(SESSION_KEY, self.random_index_str)

    def __getitem__(self, key):
        self.random_index_str = self.handler.get_secure_cookie(SESSION_KEY, None)
        if self.random_index_str:
            self.random_index_str = str(self.random_index_str, encoding='utf-8')
            return _session_manager.get_item(self.random_index_str, key)
            # current_info = _all_user_session_info.get(self.random_index_str, None)
            # if current_info:
            #     return current_info.get(key, None)
        return None

    def delete(self):
        self.random_index_str = self.handler.get_secure_cookie(SESSION_KEY, None)
        if self.random_index_str:
            self.random_index_str = str(self.random_index_str, encoding='utf-8')
            _session_manager.delete(self.random_index_str)
            # if self.random_index_str in _all_user_session_info:
            #     del _all_user_session_info[self.random_index_str]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import hashlib
import shelve
import pickle
import time
import tornado.gen

class FileCache:
    def __init__(self, cache_file):
        path_name, _ = os.path.split(cache_file)
        if not os.path.exists(path_name):
            os.makedirs(path_name, 0o777)
        self.cache = shelve.open(cache_file)

    def get_cache_key(self, key):
        hash_object = hashlib.md5(str(key).encode())
        return hash_object.hexdigest()

    @tornado.gen.coroutine
    def get(self, key):
        cache_key = self.get_cache_key(key)

        if cache_key in self.cache:
            data = self.cache[cache_key]
            if 'value' in data:
                if data.get('expiry', 0) >= time.time():
                    return data['value']
                else:
                    del self.cache[cache_key]

        return None

    @tornado.gen.coroutine
    def set(self, key, value, expire=None):
        cache_key = self.get_cache_key(key)

        if expire is not None:
            expiry_time = time.time() + expire
        else:
            expiry_time = None

        self.cache[cache_key] = {'value': value, 'expiry': expiry_time}

    @tornado.gen.coroutine
    def mget(self, keys):
        values = []

        for key in keys:
            values.append(self.get(key))

        return values

    @tornado.gen.coroutine
    def mset(self, mapping, expire=None):
        for key, value in mapping.items():
            self.set(key, value, expire)

    @tornado.gen.coroutine
    def hget(self, key, field):
        cache_key = self.get_cache_key(key)

        if cache_key in self.cache and isinstance(self.cache[cache_key], dict):
            data = self.cache[cache_key]
            if 'fields' in data:
                return data['fields'].get(field)

        return None

    @tornado.gen.coroutine
    def hset(self, key, field, value, expire=None):
        cache_key = self.get_cache_key(key)

        if cache_key not in self.cache:
            self.cache[cache_key] = {'fields': {}}

        self.cache[cache_key]['fields'][field] = value
        self.set(key, self.cache[cache_key], expire)

    @tornado.gen.coroutine
    def hmget(self, key, fields):
        cache_key = self.get_cache_key(key)

        if cache_key in self.cache and isinstance(self.cache[cache_key], dict):
            data = self.cache[cache_key]
            if 'fields' in data:
                values = []
                for field in fields:
                    values.append(data['fields'].get(field))
                return values

        return [None] * len(fields)

    @tornado.gen.coroutine
    def hmset(self, key, mapping, expire=None):
        cache_key = self.get_cache_key(key)

        if cache_key not in self.cache:
            self.cache[cache_key] = {'fields': {}}

        self.cache[cache_key]['fields'].update(mapping)
        self.set(key, self.cache[cache_key], expire)

    @tornado.gen.coroutine
    def delete(self, key):
        cache_key = self.get_cache_key(key)
        if cache_key in self.cache:
            del self.cache[cache_key]

    @tornado.gen.coroutine
    def smembers(self, key):
        cache_key = self.get_cache_key(key)

        if cache_key in self.cache:
            data = self.cache[cache_key]
            if 'members' in data:
                return data['members']

        return set()

    @tornado.gen.coroutine
    def sadd(self, key, *values):
        cache_key = self.get_cache_key(key)

        if cache_key not in self.cache:
            self.cache[cache_key] = {'members': set()}

        self.cache[cache_key]['members'].update(values)
        
    @tornado.gen.coroutine
    def hscan(self, key, cursor=0, count=None):
        cache_key = self.get_cache_key(key)

        if cache_key in self.cache:
            data = self.cache[cache_key]
            if 'fields' in data:
                fields = list(data['fields'].keys())
                cursor = int(cursor)
                if count is None:
                    count = len(fields)
                else:
                    count = int(count)

                fields_slice = fields[cursor : cursor + count]
                next_cursor = cursor + count if cursor + count < len(fields) else 0

                result = []
                for field in fields_slice:
                    result.append((field, data['fields'][field]))

                return next_cursor, result

        return 0, []

    @tornado.gen.coroutine
    def hgetall(self, key):
        cache_key = self.get_cache_key(key)

        if cache_key in self.cache and isinstance(self.cache[cache_key], dict):
            data = self.cache[cache_key]
            if 'fields' in data:
                return data['fields']

        return {}

    @tornado.gen.coroutine
    def zrange(self, key, start, stop, withscores=False):
        cache_key = self.get_cache_key(key)

        if cache_key in self.cache:
            data = self.cache[cache_key]
            if 'sorted_set' in data:
                sorted_set = data['sorted_set']

                if isinstance(sorted_set, dict):
                    sorted_set = sorted(sorted_set.items(), key=lambda x: x[1])

                if withscores:
                    return sorted_set[start : stop + 1]
                else:
                    return [item[0] for item in sorted_set[start : stop + 1]]

        return []

    @tornado.gen.coroutine
    def sismember(self, key, value):
        cache_key = self.get_cache_key(key)

        if cache_key in self.cache:
            data = self.cache[cache_key]
            if 'members' in data:
                return value in data['members']

        return False
        
    def close(self):
        self.cache.close()





#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import datetime
import tornado.gen
import aredis

from .supports import singleton
from .modelutils import model_columns, DEFAULT_SKIP_FIELDS
from .cacher.filecache import FileCache

LOG = logging.getLogger('components.cacheproxy')

@singleton
class CacheProxy(object):
    """
    Cacher agent component
    """

    def __init__(self):
        self.cache_inst = None
    
    def configure(self, conf: dict):
        if conf.get('type', None) == 'redis':
            self.configure_redis(conf)
        elif conf.get('type', None) == 'file':
            self.configure_filecache(conf.get('path', 'data/file-caching.db'))
        else:
            LOG.error('configure cacheproxy with configure:%s failed with unknown cache type', str(conf))
    
    def configure_redis(self, redis_conf: dict):
        self.cache_inst = aredis.StrictRedis(host=redis_conf.get('host', 'localhost'), port=redis_conf.get('port', 6379), 
            db=redis_conf.get('db', 0), password=redis_conf.get('password', None), retry_on_timeout=True)
        return True
    
    def configure_filecache(self, cache_path: str):
        self.cache_inst = FileCache(cache_path)

    def prepare(self):
        if self.cache_inst == None:
            self.configure_filecache('data/file-caching.db')

    @tornado.gen.coroutine
    def get_object(self, key, keys):
        self.prepare()
        res = yield self.cache_inst.hmget(key, keys)
        if not res:
            return False
        result = {}
        is_all_none = True
        i = 0
        for k in keys:
            v = res[i].decode() if isinstance(res[i], bytes) else res[i]
            result[k] = v
            i += 1
            if is_all_none and v is not None:
                is_all_none = False
        if is_all_none:
            result = False
        return result
    
    @tornado.gen.coroutine
    def get_objects(self, key, keys):
        self.prepare()
        results = []
        idxes = yield self.get_sets_values(key)
        for idx in idxes:
            row = yield self.get_object(key+':'+idx, keys)
            if row is not False:
                results.append(row)
        return results

    @tornado.gen.coroutine
    def set_object(self, key, mapping, expire=None):
        if not mapping:
            return False
        for k,v in mapping.items():
            if v is None:
                mapping[k] = ''
        yield self.cache_inst.hmset(key, mapping, expire=expire)
        return True

    @tornado.gen.coroutine
    def get_sets_values(self, key):
        results = []
        vals = yield self.cache_inst.smembers(key)
        if not vals:
            return results
        for val in vals:
            val = val.decode() if isinstance(val, bytes) else str(val)
            results.append(val)
        return results

    @tornado.gen.coroutine
    def add_sets_values(self, key, value):
        yield self.cache_inst.sadd(key, value)
        
    @tornado.gen.coroutine
    def get_sets_values_extend(self, key, keys):
        vals = yield self.get_sets_values(key)
        return self.parse_imploded_values(vals, keys)

    def parse_imploded_values(self, rows, keys):
        results = []
        if not rows:
            return results
        l = len(keys)
        for val in rows:
            one = {k:'' for k in keys}
            i = 0
            eles = val.split('-')
            l2 = len(eles)
            for i in range(l):
                if i < l2:
                    one[keys[i]] = eles[i]
            # in case that if last field is text that contains '-'
            if l > 0 and l2 > l:
                for j in range(l2-l):
                    one[keys[l-1]] += '-' + eles[l+j]
            results.append(one)
        return results
        
    @tornado.gen.coroutine
    def get_sorted_sets_values(self, key):
        results = []
        vals = yield self.zrange(key, 0, -1, withscores=True)
        if not vals:
            return results
        for ele in vals:
            val = ele[0]
            val = val.decode() if isinstance(val, bytes) else val
            results.append(val)
        return results
    
    @tornado.gen.coroutine
    def get_sorted_sets_value_extend(self, key, keys):
        vals = yield self.get_sorted_sets_values(key)
        return self.parse_imploded_values(vals, keys)

    @tornado.gen.coroutine
    def get_cache_value(self, key):
        val = yield self.cache_inst.get(key)
        return val

    @tornado.gen.coroutine
    def scan_hash_keys(self, hash_key, key_match, count=1000, cursor=0):
        res = yield self.cache_inst.hscan(hash_key, cursor, match=key_match, count=count)
        keys = []
        next_cursor = 0
        if res:
            next_cursor = res[0]
            for k in res[1]:
                keys.append(k.decode() if isinstance(k, bytes) else str(k))
        return keys, next_cursor
    
    @tornado.gen.coroutine
    def find_hash_keys(self, hash_key, key_match, match_keys = {}):
        keys = []
        if not hash_key or not key_match or not match_keys:
            return keys
        next_cursor = -1
        while next_cursor != 0:
            if next_cursor == -1:
                next_cursor = 0
            scanedKeys, next_cursor = yield self.scan_hash_keys(hash_key, key_match, cursor=next_cursor)
            for k in scanedKeys:
                if k in match_keys:
                    keys.append(k)
        return keys

    @tornado.gen.coroutine
    def get_all_hash_keys(self, hash_key_prefix, match_keys = {}):
        res = yield self.cache_inst.hgetall(hash_key_prefix)
        result = []
        for row in res:
            a = row
        return result
        
    @tornado.gen.coroutine
    def clear_by_key_prefix(self, key_prefix):
        keys = yield self.cache_inst.keys(key_prefix+'*')
        del_keys = [[]]
        i = 0
        for k in keys:
            del_keys[i].append(k.decode() if isinstance(k, bytes) else str(k))
            if len(del_keys[i]) > 50:
                i += 1
                del_keys.append([])

            if i > 10:
                yield [self.cache_inst.delete(*dkeys) for dkeys in del_keys]
                del_keys = [[]]
                i = 0
        
        if del_keys[0]:
            yield [self.cache_inst.delete(*dkeys) for dkeys in del_keys]

    @tornado.gen.coroutine
    def incr(self, key, expire = None):
        yield self.cache_inst.incr(key)
        if expire:
            yield self.cache_inst.expire(key, expire)

    @tornado.gen.coroutine
    def set(self, key, value, expire = None, px=None, nx=False, xx=False):
        """
        Set the value at key ``name`` to ``value``

        ``expire`` sets an expire flag on key ``name`` for ``expire`` seconds.

        ``px`` sets an expire flag on key ``name`` for ``px`` milliseconds.

        ``nx`` if set to True, set the value at key ``name`` to ``value`` only
            if it does not exist.

        ``xx`` if set to True, set the value at key ``name`` to ``value`` only
            if it already exists.
        """
        yield self.cache_inst.set(key, value, ex=expire, px=px, nx=nx, xx=xx)

    @tornado.gen.coroutine
    def get(self, key):
        val = yield self.cache_inst.get(key)
        return val

    @tornado.gen.coroutine
    def delete(self, key):
        val = yield self.cache_inst.delete(key)
        return val

    @tornado.gen.coroutine
    def zrange(self, name, start, end, desc=False, withscores=False,
               score_cast_func=float):
        """
        Return a range of values from sorted set ``name`` between
        ``start`` and ``end`` sorted in ascending order.

        ``start`` and ``end`` can be negative, indicating the end of the range.

        ``desc`` a boolean indicating whether to sort the results descendingly

        ``withscores`` indicates to return the scores along with the values.
        The return type is a list of (value, score) pairs

        ``score_cast_func`` a callable used to cast the score return value
        """
        val = yield self.cache_inst.zrange(name, start, end, desc=desc, withscores=withscores, score_cast_func=score_cast_func)
        return val

    @tornado.gen.coroutine
    def is_exists_in_sets(self, key, value):
        val = yield self.cache_inst.sismember(key, value)
        return val

    @tornado.gen.coroutine
    def add_to_cache_indexed_to_many(self, item, key_prefix, index_key, pk):
        if not isinstance(item, dict):
            item2 = item
            item = {}
            columns,_ = model_columns(item2)
            for k in columns:
                if k not in DEFAULT_SKIP_FIELDS:
                    item[k] = getattr(item2, k)
        pk_value = self.get_index_key_value(item, pk)
        idx_value = self.get_index_key_value(item, index_key)
        cache_key = key_prefix + idx_value
        yield self.cache_inst.sadd(cache_key, pk_value)
        cache_key += ':' + pk_value
        yield self.set_object(cache_key, item)

    @tornado.gen.coroutine
    def del_from_cache_indexed_to_many(self, item, key_prefix, index_key, pk):
        pk_value = self.get_index_key_value(item, pk)
        idx_value = self.get_index_key_value(item, index_key)
        cache_key = key_prefix + idx_value
        yield self.cache_inst.srem(cache_key, pk_value)
        cache_key += ':' + pk_value
        yield self.cache_inst.delete(cache_key)

    def get_index_key_value(self, item, index_key):
        idx_value = ''
        if isinstance(index_key, list):
            vals = []
            for k in index_key:
                if isinstance(item, dict):
                    vals.append(str(item[k]) if item[k] is not None else '')
                else:
                    vals.append(str(getattr(item, k, '')))
            idx_value = ':'.join(vals)
        else:
            if isinstance(item, dict):
                idx_value = str(item[index_key] if item[index_key] is not None else '')
            else:
                idx_value = str(getattr(item, index_key, ''))
        return idx_value



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import time
from sqlalchemy import Column, Integer, SmallInteger, String, BigInteger, func
# from sqlalchemy.ext.declarative import declarative_base
import flask
from hawthorn.utilities import get_current_timestamp_millis

def get_current_uid(*args, **kwargs):
    # uid = str(flask.session.get('uid', ''))
    uid = ''
    if flask.session:
        uid = str(flask.session.get('uid', ''))
    return uid

class AppBehevior(object):
    """
    接入方相关字段定义
    """
    app_id = Column('app_id', String(50), index=True, nullable=False)

class ModifyingBehevior(object):
    """
    修改相关字段定义
    """
    def __init__(self, session_uid=''):
        self.__session_uid = session_uid
        self.created_by.onupdate = self.get_session_uid
        self.created_by.default = self.get_session_uid
        self.updated_by.onupdate = self.get_session_uid
        self.updated_by.default = self.get_session_uid

    obsoleted = Column('obsoleted', SmallInteger, default=0, comment='废弃标志 0:正常 1:废弃')
    created_at = Column('created_at', BigInteger, default=get_current_timestamp_millis, comment='创建时间戳（毫秒）')
    updated_at = Column('updated_at', BigInteger, default=get_current_timestamp_millis, onupdate=get_current_timestamp_millis, comment='更新时间戳（毫秒）')
    created_by = Column('created_by', String(50), default=get_current_uid, comment='创建者用户ID')
    updated_by = Column('updated_by', String(50), default=get_current_uid, onupdate=get_current_uid, comment='更新者用户ID')

    def get_session_uid(self):
        if self.__session_uid:
            return self.__session_uid
        return get_current_uid()
    
    def set_session_uid(self, session_uid: str):
        self.__session_uid = str(session_uid)
        


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import redis
import logging
import datetime
import tornado.gen
import aredis

from .modelutils import model_columns, format_mongo_value, DEFAULT_SKIP_FIELDS
from .dbproxy import DbProxy
from .cacheproxy import CacheProxy

LOG = logging.getLogger('components.db2cachehelper')

@tornado.gen.coroutine
def load_mongo_data_to_cache(model, keyPrefix, pk, filters=None, excachecb=None, clearcache=True):
    cacheproxy = CacheProxy()
    check_uniques = {}
    @tornado.gen.coroutine
    def _load_cache_pk(item):
        cache_key = keyPrefix + cacheproxy.get_index_key_value(item, pk)
        if cache_key in check_uniques:
            LOG.warning('loadToCache by key:%s that already exists.', cache_key)
        check_uniques[cache_key] = 1
        yield cacheproxy.set_object(cache_key, item)
        if callable(excachecb):
            yield excachecb(item, cacheproxy.async_redis_conn)

    if clearcache:
        yield cacheproxy.clear_by_key_prefix(keyPrefix)
    yield load_data_from_mongodb(model, _load_cache_pk, filters=filters)

@tornado.gen.coroutine
def load_mongo_data_to_cache_indexed_to_many(model, keyPrefix, indexKey, pk, filters=None, orderby=None, clearcache=True):
    cacheproxy = CacheProxy()
    @tornado.gen.coroutine
    def _load_cache_index(item):
        yield cacheproxy.add_to_cache_indexed_to_many(item, keyPrefix, indexKey, pk)
    if clearcache:
        yield cacheproxy.clear_by_key_prefix(keyPrefix)
    yield load_data_from_mongodb(model, _load_cache_index, filters=filters, orderby=orderby)

@tornado.gen.coroutine
def load_data_from_mongodb(model, cb, filters=None, orderby=None):
    dbproxy = DbProxy()
    limit = 5000
    offset = 0
    nrows = limit
    modelName = str(model.__name__)
    LOG.info("loading %s from db begining", modelName)
    qfilters = []
    kwfilters = {}
    curId = None
    if isinstance(filters, tuple):
        for f in filters:
            if isinstance(f, dict):
                for k,v in f.items():
                    kwfilters[k] = v
            elif isinstance(f, list):
                for v in f:
                    qfilters.append(v)
    elif isinstance(filters, dict):
        kwfilters = filters
    elif isinstance(filters, list):
        qfilters = filters
    while nrows >= limit:
        if curId:
            kwfilters['id__gt'] = curId
        rows = yield dbproxy.query_all_mongo(model, (qfilters, kwfilters), limit)
        nrows = 0
        for row in rows:
            nrows += 1
            curId = row.get('id')
            item = {}
            for k in row:
                if k in DEFAULT_SKIP_FIELDS:
                    continue
                item[k] = format_mongo_value(row.get(k))
            if tornado.gen.is_coroutine_function(cb):
                yield cb(item)
            else:
                cb(item)
        offset += nrows
        LOG.info("loading %s from db offset:%d rows:%d", modelName, offset, nrows)
    
    LOG.info("loading %s from db finished", modelName)

